/************************************************************
* $Id: nnet.c,v 1.22 2009/08/11 15:47:44 mechanoid Exp mechanoid $
*       
*  ART1 neural net 
*
*       OS : FreeBSD 7.2
* COMPILER : GNU project C and C++ Compiler  v. 4.2.1
*
*   AUTHOR : Evgeny S. Borisov
*
*  Glushkov Institute of Cybernetics
*  National Academy of Sciences of Ukraine
* 
*    http://www.mechanoid.kiev.ua
*  e-mail : nn@mechanoid.kiev.ua
* 
************************************************************/
#include <gdk/gdk.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "nnet.h"

// vigilance parameter
#define RHO  0.8
// weight constant
#define L    2.0

data_t    layer_0[SIZE_LAYER_1];
data_t    layer_1[SIZE_LAYER_1]; 
layer_2_t layer_2[SIZE_LAYER_2];

weight_t weight_layer_1[SIZE_LAYER_2][SIZE_LAYER_1]; 
weight_t weight_layer_2[SIZE_LAYER_1][SIZE_LAYER_2];

int layer_2_winner_index;
int g;

/* - - - - - - - - - - - - - - - - - - - */
void init_weights() {
   int i,j;
   for(i=0 ; i<SIZE_LAYER_1 ; i++){
	  for(j=0 ; j<SIZE_LAYER_2 ; j++){
		 weight_layer_2[i][j]=1;
		 weight_layer_1[j][i]=( L / (L - 1.0 + SIZE_LAYER_1) );
	  }
   }
}

/* - - - - - - - - - - - - - - - - - - - */
void init_layers() {
   int i,j;
   for(i=0 ; i<SIZE_LAYER_1 ; i++){ layer_1[i]=0; }
   for(j=0 ; j<SIZE_LAYER_2 ; j++){ layer_2[j].was_winner=0; layer_2[j].state=0; layer_2[j].out=0; }
   layer_2_winner_index=0;
   g=0;
}


/* - - - - - - - - - - - - - - - - - - - */
void init_nnet() {
   int j;
   init_weights();
   for(j=0 ; j<SIZE_LAYER_2 ; j++){ layer_2[j].reserved=0; }
   init_layers();
}


/* - - - - - - - - - - - - - - - - - - - 
* calculate output of G (layer 1 control) neuron of ART-1
*/
void calc_g() {
   int i,j;
   int x,y;
   
   for(x=0, i=0 ; i<SIZE_LAYER_1 ; i++){ x += layer_0[i]; }
   for(y=0, j=0 ; j<SIZE_LAYER_2 ; j++){ y += layer_2[j].out; }

   g = x && !y;
}

/* - - - - - - - - - - - - - - - - - - - 
* calculate output layer 1 neurons
*/
void calc_layer_1_out() {
   int i;

   calc_g();
   
   for(i=0 ; i<SIZE_LAYER_1 ; i++){
	  // set neuron activity signal 
	  // if 2 from 3 input is active
	  layer_1[i]=( (layer_0[i] + g + layer_2[layer_2_winner_index].out*weight_layer_2[i][layer_2_winner_index]) > 1 ) ? 1 : 0 ; 
   }
} 
/* - - - - - - - - - - - - - - - - - - - 
* calculate state layer 2 neurons
*/
void calc_layer_2_state() {
   weight_t s;
   int i,j;
   
   for(j=0 ; j<SIZE_LAYER_2 ; j++) {
	  for(s=0, i=0 ; i<SIZE_LAYER_1 ; i++) {
		 s+=layer_1[i]*weight_layer_1[j][i];
	  }
	  layer_2[j].state=s;
   }
}

/* - - - - - - - - - - - - - - - - - - - - - 
* find layer 2 neuron with maximum activity
* from neurons "nowinners"
* return 0 if new winner found
*/
int find_layer_2_winner() {
   weight_t max_s=0;
   int max_idx=0;
   int i,j;

   for(j=0 ; j<SIZE_LAYER_2 ; j++) {
      // find first nowinner index
	  if( !(layer_2[j].was_winner) ) { // found first
		 max_idx=j;
		 max_s=layer_2[j].state;
         // find maximum 
		 for(i=j+1 ; i<SIZE_LAYER_2 ; i++) {
			if( !(layer_2[i].was_winner) ) {
			   if( layer_2[i].state > max_s ) {
				  max_s = layer_2[i].state;
				  max_idx = i;
			   }
			}
		 }
		 // fix parameters
		 layer_2[max_idx].was_winner=1;
		 layer_2_winner_index=max_idx;
		 return 0;
	  }
   }
   // not found
   return 1;
}
 
/* - - - - - - - - - - - - - - - - - - - 
* calculate output layer 2 neurons
*/
void calc_layer_2_out() {
   int j;

   for(j=0 ; j<SIZE_LAYER_2 ; j++) { layer_2[j].out=0; }
   layer_2[layer_2_winner_index].out=1;
}

/* - - - - - - - - - - - - - - - - - - - 
* check vigilance parameter,
* difference between input and backward signal vectors
* if vigilance condition is true then return 1
*/
int check_vigilance() {
   int i;
   weight_t x, y, r;

   for( x=0,y=0,i=0 ; i<SIZE_LAYER_1 ; i++ ) { x+=layer_0[i]; y+=layer_1[i]; }
   r=(x==0)?0:(y/x);
   return ( r > RHO );
}

/* - - - - - - - - - - - - - - - - - - - 
* training rules
*/
void weight_modification() {
   int i;
   weight_t y;

   for(y=0,i=0 ; i<SIZE_LAYER_1 ; i++) { y+=layer_0[i]; }
   for(i=0 ; i<SIZE_LAYER_1 ; i++) {
	  weight_layer_1[layer_2_winner_index][i]=( L*layer_0[i] / (L - 1.0 + y) );
	  weight_layer_2[i][layer_2_winner_index]=layer_0[i];
   }
}

/* - - - - - - - - - - - - - - - - - - - - -
* find first not reserved neuron in layer 2
* return 0 if found
*/
int find_layer_2_free() {
   int j;
   // check free neurons
   for( j=0 ; j<SIZE_LAYER_2 ; j++ ) { 
	  if( !(layer_2[j].reserved) ) { layer_2_winner_index=j; return 0; } 
   }
   return 1;
}

/* - - - - - - - - - - - - - - - - - - - 
* neural net main loop
*/
int step() {
   init_layers();
   calc_layer_1_out();
   calc_layer_2_state();
   find_layer_2_winner();

   for (;;) {
	  calc_layer_2_out();
	  calc_layer_1_out();
   
 
	  if(check_vigilance()) { // classification successful
		 if(layer_2[layer_2_winner_index].reserved) { 
		    printf("classification successful, cluster %d\n",layer_2_winner_index); 
			return 0;
		 } 
		 printf("new pattern, added to cluster %d\n",layer_2_winner_index);
		 layer_2[layer_2_winner_index].reserved=1;
	     // make weight modification
		 break; 
	  }
	  // vigilance check fail
      // find next winer 
	  if( find_layer_2_winner() ) { 
	     // no more winners
	     // new cluster detected
	     // find free neuron 
		 if(find_layer_2_free()) { printf("new pattern, no free cluster\n"); return 1; }
	     // make weight modification
         printf("new pattern, cluster %d reserved \n",layer_2_winner_index);
		 layer_2[layer_2_winner_index].reserved=1;
		 break;
	  }
	  
	  // found new neuron winner
	  // goto loop start, make next tests
   }
   
   // make weight modification for winner
   weight_modification();

   return 2;
}

